package tree;
public class Generate {

static java.util.LinkedList CODE = new java.util.LinkedList();

public static void Add(Tree tree) {
    CODE.addLast(tree);
}

public static Expression AssignObjectVal(
   Expression Ptr, 
   sym.TypeSy LeftType, 
   sym.TypeSy RightType, 
   int Pos) {
    if ((LeftType instanceof sym.ClassType) && 
        (RightType instanceof sym.ClassType)) {
       sym.ClassType L = (sym.ClassType)LeftType;
       sym.ClassType R = (sym.ClassType)RightType;
       if (sem.SemAux.IsProperSuperType(R, L) && !(Ptr instanceof LoadNil)) {
          LoadClassTemplate template = new LoadClassTemplate(L,Pos);
          Ptr = new NarrowVal(Ptr, L, template, Pos);
       }
    }
    return Ptr;
}

public static Designator AssignObjectVar (
   Designator Ptr, 
   sym.TypeSy LeftType, 
   sym.TypeSy RightType, 
   int Pos) {
    if ((LeftType instanceof sym.ClassType) && 
        (RightType instanceof sym.ClassType)) {
       sym.ClassType L = (sym.ClassType)LeftType;
       sym.ClassType R = (sym.ClassType)RightType;
       if (sem.SemAux.IsProperSuperType(R, L)) {
          LoadClassTemplate template = new LoadClassTemplate(L,Pos);
          Ptr = new NarrowVar(Ptr, L, template, Pos);
       }
    }
    return Ptr;
}

public static void PurgeLabels() {
   java.util.HashSet referenced = new java.util.HashSet();
   java.util.ListIterator i = CODE.listIterator();
   while (i.hasNext()) {
      Tree t = (Tree) i.next();
      if (t instanceof Control) {
         Control c = (Control) t;
         referenced.add(c.label);
      }
   }

   java.util.HashSet alreadyGenerated = new java.util.HashSet();
   java.util.ListIterator j = CODE.listIterator();
   while (j.hasNext()) {
      Tree t = (Tree) j.next();
      if (t instanceof Label) {
         Label c = (Label) t;
         if (! referenced.contains(c.label))
            j.remove();
         else if (alreadyGenerated.contains(c.label))
            j.remove();
         else
            alreadyGenerated.add(c.label);
      }
   }
}

//---------------------------------------------------------
//             Perform Intermediate Code Generation
//---------------------------------------------------------
public static java.util.LinkedList GenTreeCode(ast.PROGRAM E) {
    Add(new Version(aux.Version.MAJOR, aux.Version.MINOR, E.position));

    Declarations(E.decls);
    java.util.LinkedList procs = ListProcedures(E.decls, new java.util.LinkedList());
    java.util.ListIterator i = procs.listIterator();
    while (i.hasNext()) {
       ast.DECLARATION p = (ast.DECLARATION) i.next();
       GenProcedure(p);
    }

    GenVars(sym.Standard.MainProc);
    Add(new ProcBegin(sym.Standard.MainProc, E.position));
    GenStatements(E.stats);
    Add(new Label(E.label,E.position));
    Add(new ProcEnd(sym.Standard.MainProc, E.position));

    PurgeLabels();
    return CODE;
}

public static void GenVars(sym.ProcedureSy proc) {
   sym.SyTab locals = proc.GetProcLocals();
   java.util.Iterator i = locals.iterator();
   while (i.hasNext()) {
      sym.VariableSy v = (sym.VariableSy) i.next();
      Add(new VarDecl(v, v.GetPos()));
   }  
}

public static void GenFormalCopies(sym.ProcedureSy proc) {
   sym.SyTab formals = proc.GetProcFormals();
   java.util.Iterator i = formals.iterator();
   while (i.hasNext()) {
      sym.FormalSy f = (sym.FormalSy) i.next();
      sym.VariableSy copy = f.GetFormalCopy();
      if (copy != null) {
         int pos = f.GetPos();
         sym.TypeSy Type = f.GetType();
         Designator ToDes = new AddressOf(copy, Type, pos);
         Designator FromAddr = new AddressOf(f, Type, pos);
         Load FromDes= new Load(Type, FromAddr, pos);
         Add(new Copy(Type, ToDes, FromDes, pos));
      }
   }  
}

public static java.util.LinkedList ListProcedures(ast.AST E, java.util.LinkedList In) {
      if (E instanceof ast.DECLNULL) {
         return In;
      } else if (E instanceof ast.CLASSDECL) {
         ast.CLASSDECL e = (ast.CLASSDECL) E;
          return ListProcedures(e.methods, In);
      } else if (E instanceof ast.METHODDECL) {
         ast.METHODDECL e = (ast.METHODDECL) E;
          In.addLast(e);
          return ListProcedures(e.decls, In);
      } else if (E instanceof ast.PROCDECL) {
          ast.PROCDECL e = (ast.PROCDECL) E;
          In.addLast(e);
          return ListProcedures(e.decls, In);
      } else if (E instanceof ast.DECLS) {
         ast.DECLS e = (ast.DECLS) E;
         java.util.LinkedList Out = ListProcedures(e.left, In);
         return ListProcedures(e.right, Out);
      } else return In;
}

public static void GenProcedure(ast.DECLARATION E) {
   if (E instanceof ast.PROCDECL) {
      ast.PROCDECL e = (ast.PROCDECL) E;
      sym.ProcedureSy symbol = (sym.ProcedureSy) e.symbol;
      Add(new ProcBegin(symbol, e.position));
      Declarations(e.formals);
      GenVars(symbol);
      Declarations(e.decls);
      GenFormalCopies(symbol);
      GenStatements(e.stats);
      Add(new Label(e.label,e.position));
      Add(new ProcEnd(symbol,e.position));
   } else{
      ast.METHODDECL e = (ast.METHODDECL) E;
      sym.ProcedureSy symbol = ((sym.MethodSy)E.symbol).GetMethodProc();
      Add(new ProcBegin(symbol, e.position));
      sym.ProcedureSy mySymbol = ((sym.MethodSy)E.symbol).GetMethodProc();
      sym.SyTab formals = mySymbol.GetProcFormals();  
      sym.FormalSy Self  = (sym.FormalSy) formals.locateByName("SELF");
      Add(new FormalDecl(Self,e.position));
      Declarations(e.formals);
      GenVars(symbol);
      Declarations(e.decls);
      GenFormalCopies(symbol);
      GenStatements(e.stats);
      Add(new Label(e.label,e.position));
      Add(new ProcEnd(symbol,e.position));
   }
}

public static void Declarations(ast.AST E) {
          if (E instanceof ast.DECLNULL) { 
	      //   } else if (E instanceof ast.VARDECL) { 
	      //       ast.VARDECL e = (ast.VARDECL)E;
	      //       sym.VariableSy f = (sym.VariableSy)e.symbol;
	      //       Add(new VarDecl(f,e.position));
   } else if (E instanceof ast.FORMALDECL) {
       ast.FORMALDECL e = (ast.FORMALDECL)E;
       sym.FormalSy f = (sym.FormalSy)e.symbol;
       Add(new FormalDecl(f,e.position));
   } else if (E instanceof ast.RECORDDECL) {
       ast.RECORDDECL e = (ast.RECORDDECL)E;
       sym.RecordType r = (sym.RecordType) e.symbol;
       Add(new TypeDecl(r,e.position));
   } else if (E instanceof ast.ARRAYDECL) {
       ast.ARRAYDECL e = (ast.ARRAYDECL)E;
       sym.ArrayType a = (sym.ArrayType)e.symbol;
       Add(new TypeDecl(a,e.position));
   } else if (E instanceof ast.CLASSDECL) {
       ast.CLASSDECL e = (ast.CLASSDECL)E;
       sym.ClassType c = (sym.ClassType)e.symbol;
       Add(new TypeDecl(c,e.position));
       Declarations(e.methods);
   }  else if (E instanceof ast.DECLS) {
       ast.DECLS e = (ast.DECLS)E;
       Declarations(e.left);
       Declarations(e.right);
   }
}
   
public static void GenStatements (ast.AST E) {
   if (E instanceof ast.STATNULL) {
      ast.STATNULL e = (ast.STATNULL)E;
      Add(new Label(e.label,e.position));
   } else if (E instanceof ast.ASSIGN) {
     ast.ASSIGN e = (ast.ASSIGN)E;
     if (e.left.type == sym.Standard.BoolType) {
        GenBoolExpression(e.right);
        Add(new Label(e.right.true_,e.right.position));
        Designator Left1 = GenDesignator(e.left, null);
        Expression Right1 = new LoadLit(
                       sym.Standard.IntType,
                       new sym.IntValue(sym.Standard.TrueSy.GetEnumValue()), 
                       e.position);
            Add(new Store(sym.Standard.IntType, Left1, Right1, e.position));
            Add(new Goto(e.next,e.position));
            Add(new Label(e.right.false_,e.right.position));
            Designator Left2 = GenDesignator(e.left, null);
            Expression Right2 = new LoadLit(
                       sym.Standard.IntType, 
                       new sym.IntValue(sym.Standard.FalseSy.GetEnumValue()), 
                       e.position);
            Add(new Store(sym.Standard.IntType, Left2, Right2, e.position));
          } else {
             Designator Left = GenDesignator(e.left, null);
             Expression Right = GenExpression(e.right);
             Right = AssignObjectVal(Right, e.left.type, e.right.type, e.position);
             Add(new Store(e.left.type, Left, Right, e.position));
	 }
      } else if (E instanceof ast.PROCCALL) {
        ast.PROCCALL e = (ast.PROCCALL)E;
        Actual actuals = GenActuals(e.actuals);
        if (e.symbol instanceof sym.MethodSy) {
           Designator Symbol = GenDesignator(e.des, null);
           sym.MethodSy m = (sym.MethodSy) e.symbol;
           sym.ProcedureSy p = m.GetMethodProc();
           sym.SyTab formals = p.GetProcFormals();
           sym.FormalSy Self   = (sym.FormalSy)formals.locateByName("SELF");
           Add(new MethodCall(m, Self, Symbol, actuals, e.position));
        } else {
           sym.ProcedureSy p = (sym.ProcedureSy) e.symbol;
           Add(new ProcCall(p, actuals, e.position));
        }
      } else if (E instanceof ast.WRITE) {
       ast.WRITE e = (ast.WRITE)E;
          Expression Left = GenExpression(e.expr);
          Add(new Write(e.expr.type, Left, e.position));
      } else if (E instanceof ast.SPECIAL) {
          ast.SPECIAL e = (ast.SPECIAL)E;
          Add(new Special(e.value, e.position));
      } else if (E instanceof ast.WRITELN) {
       ast.WRITELN e = (ast.WRITELN)E;
          Add(new WriteLn(e.position));
      } else if (E instanceof ast.READ) {
       ast.READ e = (ast.READ)E;
          Designator Left = GenDesignator(e.des, null);
          Add(new Read(e.des.type, Left, e.position));
      } else if (E instanceof ast.WHILE) {
       ast.WHILE e = (ast.WHILE)E;
          GenBoolExpression(e.expr);
          Add(new Label(e.expr.true_, e.position));
          GenStatements(e.stats);
          Add(new Goto(e.label, e.position));
      } else if (E instanceof ast.BLOCK) {
       ast.BLOCK e = (ast.BLOCK)E;
          Declarations(e.decls);
          GenStatements(e.stats);
      } else if (E instanceof ast.REPEAT) {
       ast.REPEAT e = (ast.REPEAT)E;
          GenStatements(e.stats);
          Add(new Label(e.begin, e.position));
          GenBoolExpression(e.expr);
      } else if (E instanceof ast.LOOP) {
       ast.LOOP e = (ast.LOOP)E;
          Add(new Label(e.begin, e.position));
          GenStatements(e.stats);
          Add(new Goto(e.begin, e.position));
      } else if (E instanceof ast.EXIT) {
       ast.EXIT e = (ast.EXIT)E;
          Add(new Goto(e.currentLoopExit, e.position));
      } else if (E instanceof ast.IF1) {
       ast.IF1 e = (ast.IF1)E;
          GenBoolExpression(e.expr);
          Add(new Label(e.expr.true_, e.position));
          GenStatements(e.then_);
      } else if (E instanceof ast.IF2) {
       ast.IF2 e = (ast.IF2)E;
          GenBoolExpression(e.expr);
          Add(new Label(e.expr.true_, e.position));
          GenStatements(e.then_);
          Add(new Goto(e.next, e.position));
          Add(new Label(e.expr.false_, e.position));
          GenStatements(e.else_);
      } else if (E instanceof ast.FOR) {
          ast.FOR e = (ast.FOR)E;
          int by  = ((sym.IntValue)e.byValue).getValue();
          int op = (by >= 0)?lexer.Token.LT:lexer.Token.GT;
          sym.VariableSy v = (sym.VariableSy) e.symbol;
          sym.TypeSy Type = v.GetType();
          Expression From   = GenExpression(e.from);
          Add(new Store(
                Type, 
                new AddressOf(v, Type, e.position), 
                From, 
                e.position));

          Expression To = GenExpression(e.to);
          Add(new Store(
                   Type, 
                   new AddressOf(e.tempTo, Type, e.position),
                   To, 
                   e.position));

          Add(new Label(e.forLabel, e.position));

          Add(new Branch(
                op, 
                Type, 
                new Load(
                       Type, 
                       new AddressOf(e.tempTo, Type, e.position), 
                       e.position), 
                new Load(
                       Type, 
                       new AddressOf(v, Type, e.position), 
                       e.position), 
                e.endLabel,
                e.position));

          GenStatements(e.stats);

          Add(new Label(e.incrLabel, e.position));

          Add(new Store(
                Type, 
                new AddressOf(v, Type, e.position),
                new BinExpr(lexer.Token.PLUS, 
                    Type,
                    new Load(
                       Type, 
                       new AddressOf(v, Type, e.position), 
                       e.position), 
                    new LoadLit(Type, new sym.IntValue(by), e.position),
                    e.position),
                e.position));

          Add(new Goto(e.forLabel, e.position));

          Add(new Label(e.endLabel, e.position));
      } else if (E instanceof ast.STATS) {
      ast.STATS e = (ast.STATS)E;
      Add(new Label(e.label,e.position));
      Add(new Label(e.left.label,e.position));
      GenStatements(e.left);
      GenStatements(e.right);
   }
}

public static Designator GenDesignator (ast.DESIGNATOR E, Designator BaseAddress) {
     Designator A = null;
     if (E instanceof ast.VARREF) {
         ast.VARREF e = (ast.VARREF) E;
//         if (e.symbol instanceof sym.ProcedureSy)   FIXME!
//             return e.symbol;

          if (e.symbol instanceof sym.FormalSy) {
             sym.FormalSy f = (sym.FormalSy) e.symbol;
             sym.VariableSy copy = f.GetFormalCopy();
             if (copy != null)
                e.symbol = copy;
          }

          if (e.symbol instanceof sym.FieldSy) {
             sym.FieldSy f = (sym.FieldSy) e.symbol;
             if (f.GetParent() instanceof sym.ClassType) {
                sym.FormalSy Self = (sym.FormalSy) e.env.locateByName("SELF");
                sym.ClassType Class = (sym.ClassType) Self.GetType();
                AddressOf SelfAddr = new AddressOf(Self, Class, e.position);
                A = new Load(Class, SelfAddr, e.position);
                A = new ObjectFieldOf(Class, f, Class, A, e.position);
             }
          } else if (e.symbol instanceof sym.MethodSy) {
             sym.MethodSy f = (sym.MethodSy) e.symbol;
             if (f.GetParent() instanceof sym.ClassType) {
                sym.FormalSy Self = (sym.FormalSy) e.env.locateByName("SELF");
                sym.ClassType Class = (sym.ClassType)Self.GetType();
                A =  new Load(
                             Class, 
                             new AddressOf(Self, Class, e.position), 
                             e.position);
             }
	  } else {
             sym.VariableSy v = (sym.VariableSy) e.symbol;
             A = new AddressOf(v, e.next.typeIn, e.position);
             if (v instanceof sym.FormalSy) {
                sym.FormalSy f = (sym.FormalSy) v;
                String mode = f.GetFormalMode();
                if (mode.equals("VAR"))
                   A = new Load(sym.Standard.IntType, A, e.position);
             }
          }
    } else if (E instanceof ast.DESNULL) {
         ast.DESNULL e = (ast.DESNULL) E;
         return BaseAddress;
    } else if (E instanceof ast.INDEX) {
       ast.INDEX e = (ast.INDEX) E;
       Expression IdxExpr = GenExpression(e.index);
       sym.ArrayType arr = (sym.ArrayType) e.typeIn;
       A = new IndexOf(arr, e.next.typeIn, BaseAddress, IdxExpr, e.position);
    } else if (E instanceof ast.FIELDREF) {
       ast.FIELDREF e = (ast.FIELDREF) E;
       sym.RecordType rec = (sym.RecordType)e.typeIn;
       sym.FieldSy field = (sym.FieldSy) e.symbol;
       A = new FieldOf(rec, field, e.next.typeIn, BaseAddress,  e.position);
    } else if (E instanceof ast.CLASSREF) {
       ast.CLASSREF e = (ast.CLASSREF) E;
       A =  new Load(e.typeIn, BaseAddress, e.position);
       if (e.symbol instanceof sym.FieldSy) {
          sym.FieldSy field = (sym.FieldSy)e.symbol;
          sym.TypeSy type = field.GetType();
          sym.ClassType class_ = (sym.ClassType)e.typeIn;
          A = new ObjectFieldOf(class_, field, type, A, e.position);
       }
    } else if (E instanceof ast.NARROW) {
       ast.NARROW e = (ast.NARROW) E;
       sym.ClassType class_ = (sym.ClassType) e.next.typeIn; 
       LoadClassTemplate template = new LoadClassTemplate(class_, e.position);
       A = new NarrowVar(BaseAddress, class_, template, e.position);
   }
   return GenDesignator(E.next, A);
}

public static Actual GenActuals (ast.ACTUAL E) {
    Actual A = null;
   if (E instanceof ast.ACTUALNULL) {
       A = new Null(E.position);
   } else if (E instanceof ast.ACTUAL) {
       ast.ACTUAL e = (ast.ACTUAL) E;
       sym.FormalSy f = e.formal;
       String mode = f.GetFormalMode();
       System.out.println(f);
       System.out.println(e.toString(0));
       sym.TypeSy FormalType = f.GetType();
       if (mode.equals("VAR") || f.GetSize() > sym.Standard.IntType.GetSize()) {
           Designator Des = GenDesignator((ast.DESIGNATOR)e.expr, null);
           Des = AssignObjectVar(Des, FormalType, e.expr.type, e.position);
           Actual next = GenActuals(e.nextActual);
           A = new Actual(FormalType, e.formal, Des, next, e.position);
       } else {
           Expression Expr = GenExpression(e.expr);
           Expr = AssignObjectVal(Expr, FormalType, e.expr.type, e.position);
           Actual next = GenActuals(e.nextActual);
           A = new Actual(FormalType, e.formal, Expr, next, e.position);
       }
   }
   return A;
}

public static void GenBoolExpression(ast.EXPRESSION E) {
   if (E instanceof ast.VARREF) {
       ast.VARREF e = (ast.VARREF) E;
       Expression Left  = GenExpression(E);
       Left.type = sym.Standard.IntType;   // Non-working hack. Should fail for (x=TRUE)=FALSE.
       sym.Value trueVal = new sym.IntValue(sym.Standard.TrueSy.GetEnumValue());
       Expression Right = new LoadLit(sym.Standard.IntType, trueVal, e.position);
       Add(new Branch(lexer.Token.EQ, sym.Standard.IntType, Left, Right, e.true_, e.position));
       Add(new Goto(e.false_, e.position));
      } else if (E instanceof ast.BINARY) {
          ast.BINARY e = (ast.BINARY) E;
          switch (e.op) {
             case lexer.Token.LT:
             case lexer.Token.GT:
             case lexer.Token.EQ:
             case lexer.Token.NE:
             case lexer.Token.LE:
             case lexer.Token.GE:
                  Expression Left = GenExpression(e.left);
                  Expression Right = GenExpression(e.right);
                  Add(new Branch(e.op, e.left.type, Left, Right, e.true_, e.position));
                  Add(new Goto(e.false_, e.position));
             break;
             case lexer.Token.AND:
                  GenBoolExpression(e.left);
                  Add(new Label(e.left.true_, e.position));
                  GenBoolExpression(e.right);
             break;
             case lexer.Token.OR:
                  GenBoolExpression(e.left);
                  Add(new Label(e.left.false_, e.position));
                  GenBoolExpression(e.right);
             }
      } else if (E instanceof ast.ISA) {
          ast.ISA e = (ast.ISA)E;
          Expression Left = GenExpression(e.left);
          sym.ClassType class_ = (sym.ClassType) e.symbol;
          Expression Right = new LoadClassTemplate(class_, e.position);
          Add(new Branch(lexer.Token.ISA, e.left.type, Left, Right, e.true_, e.position));
          Add(new Goto(e.false_, e.position));
      } else if (E instanceof ast.UNARY) {
          ast.UNARY e = (ast.UNARY)E;
          if (e.op == lexer.Token.NOT) {
              GenBoolExpression(e.left);
          }
    }
}

public static Expression GenExpression(ast.EXPRESSION E) {
   Expression A = null;
   if (E instanceof ast.ENUMREF) {
      ast.ENUMREF e = (ast.ENUMREF) E;
      sym.TypeSy Type = e.type;
      sym.EnumSy eid = (sym.EnumSy) e.symbol;
      sym.IntValue eval = new sym.IntValue(eid.GetEnumValue());
      A = new LoadLit(Type, eval, e.position);
   } else if (E instanceof ast.VARREF) {
      ast.VARREF e = (ast.VARREF) E;
      sym.TypeSy Type = e.type;
      if (e.symbol instanceof sym.ConstSy) {
         sym.ConstSy c = (sym.ConstSy) e.symbol;
         if (c == sym.Standard.NilSy) 
             A = new LoadNil(Type, e.position);
         else {
            sym.Value val = c.GetConstantValue();
            A = new LoadLit(Type, val, e.position);
         } 
      } else {
         Designator Des = GenDesignator(e, null);
        A = new Load(Type, Des, e.position);
      }
   } else if (E instanceof ast.LITERAL) {
      ast.LITERAL e = (ast.LITERAL) E;
      sym.TypeSy Type = e.type;
      A = new LoadLit(e.type, e.value, e.position);
   } else if (E instanceof ast.BINARY) {
      ast.BINARY e = (ast.BINARY) E;
      Expression Left = GenExpression(e.left);
      Expression Right = GenExpression(e.right);
      A = new BinExpr(e.op, e.type, Left, Right, e.position);
   } else if (E instanceof ast.UNARY) {
      ast.UNARY e = (ast.UNARY) E;
      Expression Left = GenExpression(e.left);
      A = new UnaryExpr(e.type, e.op, Left, e.position);
   } else if (E instanceof ast.NEW) {
      ast.NEW e = (ast.NEW) E;
      sym.ClassType c = (sym.ClassType)e.symbol;
      LoadClassTemplate template = new LoadClassTemplate(c, e.position);
      A = new UnaryExpr(e.type, lexer.Token.NEW, template, e.position);
  }
  return A;
}
}